import asyncio
import curses

from .cli import progress, success, warning


def init_curses(num_tasks):
    stdscr = curses.initscr()
    height, width = stdscr.getmaxyx()
    win = curses.newpad(num_tasks, 300)
    curses.start_color()
    curses.use_default_colors()
    # -1 is default terminal background color
    curses.init_pair(1, curses.COLOR_RED, -1)
    curses.init_pair(2, curses.COLOR_GREEN, -1)
    curses.noecho()
    curses.cbreak()
    # Turn off cursor
    curses.curs_set(0)
    return (win, height, width)


def addstr(win_info, x, y, text, color=0):
    win, height, width = win_info
    win.addstr(x, y, text, color)
    win.clrtoeol()
    win.clearok(1)
    win.refresh(0, 0, 0, 0, height - 1, width - 1)


async def run_tasks_async_with_progress(tasks, max_concurrency=10):
    sem = asyncio.Semaphore(max_concurrency)
    win_info = init_curses(len(tasks) + 1)
    addstr(
        win_info,
        0,
        0,
        f"Your command is now running on the following {len(tasks)} servers (may extend off bottom"
        " of terminal):",
    )
    # Ugly.
    just = len(max(tasks, key=lambda x: len(x[1]))[1]) + 1
    tasks = [
        print_async_complete(task, num + 1, just, win_info, sem) for num, task in enumerate(tasks)
    ]
    # Can reverse tasks with [::-1] if we prefer them to run bottom to top on output.
    outputs = await asyncio.gather(*tasks, return_exceptions=True)
    addstr(win_info, 0, 0, "Done. Press enter to print the outputs of the commands.")
    curses.echo()
    curses.nocbreak()
    input()
    curses.endwin()
    print("\n".join(outputs))


async def print_async_complete(task, position, just, win_info, sem):
    cor, name = task
    addstr(win_info, position, 0, name)
    output = f'{"-"*20}\n{progress(name)}\n{"-"*20}\n'
    try:
        await sem.acquire()
        output += await cor
    except Exception as e:
        addstr(win_info, position, just, "✗", curses.color_pair(1))
        output += f"Exception: {str(e)}"
    else:
        addstr(win_info, position, just, "✔", curses.color_pair(2))
    sem.release()
    return output


async def async_run(cmd):
    proc = await asyncio.create_subprocess_shell(
        cmd, stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.PIPE
    )

    stdout, stderr = await proc.communicate()
    return (
        stdout,
        stderr,
        (
            success("[stdout]\n")
            + f"{stdout.decode()}\n"
            + warning("[stderr]\n")
            + f"{stderr.decode()}\n"
        ),
    )



import os
from pathlib import Path
from typing import List, Union

from colorama import Fore
from tabulate import tabulate

from gitops.common.app import DEPLOYMENT_ATTRIBUTES, App
from gitops.utils import yaml as yaml

from . import get_account_id
from .cli import colourise, confirm, warning
from .exceptions import AppDoesNotExist, AppOperationAborted
from .images import colour_image
from .tags import colour_tags, validate_tags

APPS_PATH = Path(os.environ.get("GITOPS_APPS_DIRECTORY", "apps"))


def get_app_details(app_name: str, load_secrets: bool = True) -> App:
    account_id = get_account_id() if load_secrets else "UNKNOWN"
    try:
        app = App(
            app_name, path=APPS_PATH / app_name, load_secrets=load_secrets, account_id=account_id
        )
    except FileNotFoundError:
        # Check if apps dir doesn't exist, or just that one app
        if os.path.exists("apps"):
            raise AppDoesNotExist(app_name)
        else:
            raise AppDoesNotExist()

    return app


def update_app(app_name: str, **kwargs):
    filename = APPS_PATH / app_name / "deployment.yml"
    with open(filename, "r") as f:
        data = yaml.safe_load(f)
    for k, v in kwargs.items():
        if k not in DEPLOYMENT_ATTRIBUTES:
            print(warning(f"Key '{k}' is not a recognised deployment attribute for {app_name}."))
        if v in [[], {}]:
            if k in data:
                del data[k]
        else:
            data[k] = v
    with open(filename, "w") as f:
        yaml.dump(data, f, default_flow_style=False)


def get_apps(
    filter: Union[List[str], str] = "",
    exclude: Union[List[str], str] = "",
    mode="PROMPT",
    autoexclude_inactive=True,
    message=None,
    load_secrets=True,
) -> List[App]:
    """Return apps that contain ALL of the tags listed in `filter` and NONE of the tags listed in
    `exclude`. The incoming filter and exclude params may come in as a list or commastring.
    For the purpose of this filtering, app names and image tag prefixes are also considered as
    tags. For instance, you can do get_apps(tags=[emeriss, production], exclude=[arafire]).
    Calling this method without any args returns all apps.
    There are three modes for communicating selected apps to the user:
    - PROMPT: Prints selected apps and asks for confirmation to proceed.
    - PREVIEW: Prints selected apps then proceeds.
    - SILENT: Proceeds without printing.
    Apps with the `inactive` tag are excluded by default, unless requested otherwise.
    """
    if filter == "all":
        filter = set()
    else:
        filter = set(filter.split(",") if filter and isinstance(filter, str) else filter)

    exclude = set(exclude.split(",") if exclude and isinstance(exclude, str) else exclude)

    if autoexclude_inactive:
        exclude.add("inactive")

    apps = []
    existing_tags = set()

    try:
        directory = sorted(APPS_PATH.iterdir())
    except FileNotFoundError:
        raise AppDoesNotExist()
    for entry in directory:
        if not entry.is_dir():
            continue
        app = get_app_details(entry.name, load_secrets=load_secrets)

        pseudotags = [app.name, app.cluster]
        if app.image:
            pseudotags.append(app.image_prefix)

        tags = set(app.tags + pseudotags)
        existing_tags |= tags
        if filter <= tags and not exclude & tags:
            apps.append(app)

    validate_tags(filter | exclude, existing_tags)

    if mode in ["PROMPT", "PREVIEW"]:
        if mode == "PROMPT" and message is None:
            message = "The following apps will be affected:"
        if message is not None:
            print(colourise(f"{message}\n", Fore.LIGHTBLUE_EX))

        preview_apps(apps)

        if mode == "PROMPT" and not confirm():
            raise AppOperationAborted

    return apps


def preview_apps(apps: List[App]):
    """Produce a summary of apps, their tags, and their expected images & replicas.
    May not necessarily reflect actual app statuses if recent changes haven't yet been pushed to
    the remote, or the deployment has failed.
    """
    table = []
    for app in apps:
        table.append(
            [
                colourise(app.name, Fore.RED, lambda _: "inactive" in app.tags),
                colour_image(app.image_tag),
                app.cluster,
                colourise(
                    app.values.get("containers", {}).get("fg", {}).get("replicas", "-"),
                    Fore.LIGHTBLACK_EX,
                    lambda r: r == "-",
                ),
                colourise(
                    app.values.get("containers", {}).get("bg", {}).get("replicas", "-"),
                    Fore.LIGHTBLACK_EX,
                    lambda r: r == "-",
                ),
                colour_tags(app.tags),
            ]
        )
    # Sort table by app tags. TODO: Do we want this over alphabetical app sorting..?
    # table = sorted(table, key=lambda x: x[4])
    print(tabulate(table, ["Name", "Image", "Cluster", "FGs", "BGs", "Tags"]))



from colorama import Fore


def colourise(value, colour, condition=None):
    """Colour a piece of text. If a condition callback is passed in, the text will only be
    coloured if the condition is met.
    """
    if condition is not None:
        if not condition(value):
            return str(value)
    return f"{colour}{value}{Fore.RESET}"


def success(text):
    return colourise(text, Fore.GREEN)


def success_negative(text):
    return colourise(text, Fore.LIGHTRED_EX)


def progress(text):
    return colourise(text, Fore.CYAN)


def warning(text):
    return colourise(text, Fore.RED)


def confirm(message=""):
    prompt_message = colourise("\nDo you want to continue? [y/N] ", Fore.LIGHTBLUE_EX)
    response = input(f"{message}{prompt_message}")
    return response.lower() in ("y", "yes")



import pkg_resources
from invoke import Collection, Program, Task

from . import core, db, shorthands

version = pkg_resources.require("gitops")[0].version

namespace = Collection()

# Load up some of our functions into the root namespace.
for core_ns in [core, shorthands]:
    tasks = filter(lambda x: isinstance(x, Task), vars(core_ns).values())
    for task in tasks:
        namespace.add_task(task)

# Namespace the rarer ones.
namespace.add_collection(db)

program = Program(namespace=namespace, version=version)



import os
from typing import Dict

import yaml


def load_yaml(path) -> Dict:
    with open(path, "r") as file:
        return resolve_values(yaml.safe_load(file), path)


def resolve_values(values, path) -> Dict:
    if "extends" not in values:
        return values
    parent_values = load_yaml(os.path.join(os.path.dirname(path), values["extends"]))
    return deep_merge(parent_values, values)


def deep_merge(parent: Dict, child: Dict) -> Dict:
    """Deeply merge two dictionaries.

    Dictionary entries will be followed and merged, anything else will be
    replaced. If the child dictionary has overlapping values. `child` is merged
    into `parent`. The operation is in-place, but the result is still returned.
    """
    for key, value in child.items():
        parent_value = parent.get(key)
        if isinstance(parent_value, dict):
            if isinstance(value, dict):
                deep_merge(parent_value, value)
            else:
                parent[key] = value
        else:
            parent[key] = value
    return parent



#############
# This code has been pinched from: https://github.com/wimglenn/oyaml/blob/master/oyaml.py
# It tweaks the yaml package to make it preserve ordering when dumping dicts.
#############
from collections import OrderedDict

import yaml as pyyaml


def map_representer(dumper, data):
    return dumper.represent_dict(data.items())


def map_constructor(loader, node):
    loader.flatten_mapping(node)
    return OrderedDict(loader.construct_pairs(node))


# This monkeypatch ensures we get nice indentation for dashed lists.
def increase_indent(self, flow=False, indentless=False):
    return super(pyyaml.dumper.Dumper, self).increase_indent(flow=flow, indentless=False)


if pyyaml.safe_dump is pyyaml.dump:
    # PyYAML v4.1
    SafeDumper = pyyaml.dumper.Dumper
    DangerDumper = pyyaml.dumper.DangerDumper
else:
    SafeDumper = pyyaml.dumper.SafeDumper
    DangerDumper = pyyaml.dumper.Dumper

SafeDumper.increase_indent = increase_indent
DangerDumper.increase_indent = increase_indent

pyyaml.add_representer(dict, map_representer, Dumper=SafeDumper)
pyyaml.add_representer(OrderedDict, map_representer, Dumper=SafeDumper)
pyyaml.add_representer(dict, map_representer, Dumper=DangerDumper)
pyyaml.add_representer(OrderedDict, map_representer, Dumper=DangerDumper)

del map_constructor, map_representer


# Merge PyYAML namespace into ours.
# This allows users a drop-in replacement:
#   import utils.yaml as yaml
from yaml import *  # noqa isort:skip



import asyncio
import inspect
import os
import random
import shlex
import string
import sys
import tempfile
import textwrap
import time
from base64 import b64encode
from contextlib import contextmanager
from datetime import datetime
from functools import wraps
from typing import Dict

import boto3
import humanize
import yaml
from colorama import Fore
from invoke import run
from invoke.exceptions import UnexpectedExit

from gitops.common.app import App
from gitops.utils.apps import APPS_PATH
from gitops.utils.async_runner import async_run

from .exceptions import CommandError


async def run_job(app: App, command, cleanup=True, sequential=True, fargate=False):
    job_id = make_key(4).lower()
    values = {
        "name": f"{app.name}-command-{job_id}",
        "app": app.name,
        "command": str(shlex.split(command)),
        "image": app.image,
    }
    extra_labels = {}
    if fargate:
        extra_labels = {"uptick/fargate": "true"}
    return await _run_job(
        "jobs/command-job.yml",
        values,
        context=app.cluster,
        namespace="workforce",
        attach=True,
        cleanup=cleanup,
        sequential=sequential,
        extra_labels=extra_labels,
    )


def list_backups(product, prefix):
    """List application backups."""
    backups = get_backups(product, prefix)
    for ii, backup in enumerate(reversed(backups)):
        index = len(backups) - ii
        size = humanize.naturalsize(backup[2])
        print(f"{Fore.BLUE}{index:4}{Fore.RESET}. {backup[1]} - {Fore.GREEN}{size}{Fore.RESET}")


def download_backup(product, prefix, index, path=None, datestamp=False):
    s3 = boto3.client("s3")
    key = get_backups(product, prefix)[int(index) - 1][3]
    url = s3.generate_presigned_url("get_object", Params={"Bucket": "uptick-backups", "Key": key})
    name = prefix
    if datestamp:
        name += f"-{get_backup_datestamp(url)}"
    name += ".dump"
    if path:
        name = os.path.join(path, name)
    print(f"Downloading to {Fore.BLUE}{name}{Fore.RESET} ... ", end="", flush=True)
    run(f'curl "{url}" -s | zcat > {name}')
    print(f"{Fore.GREEN}ok{Fore.RESET}")


def copy_db(ctx, source, destination, context=""):
    """Copy database between two applications."""
    with tempfile.NamedTemporaryFile("wt", suffix=".yml") as tmp:
        job = open("copy-db-job.yml", "r").read()
        job = job.replace("{{ source }}", source)
        job = job.replace("{{ destination }}", destination)
        tmp.write(job)
        tmp.flush()
        run(f"kubectl create -n workforce -f {tmp.name}", hide=True)
        print("Waiting for copy to complete ... ", end="", flush=True)
        run(
            (
                "kubectl wait"
                f" --context {context}"
                " -n workforce"
                " --for=condition=complete"
                " --timeout=-1s"
                " job/copy-db-{source}-{destination}"
            ).format(source=source, destination=destination),
            hide=True,
        )
        print("done")
        run(
            ("kubectl delete -n workforce job copy-db-{source}-{destination}").format(
                source=source, destination=destination
            ),
            hide=True,
        )


def get_backups(product, prefix):
    s3 = boto3.resource("s3")
    bucket = s3.Bucket("uptick-backups")
    all_backups = []
    for obj in bucket.objects.filter(Prefix=f"{product}/{prefix}/"):
        name = obj.key.split("/")[-1].split(".")[0]
        if not name:
            continue
        dt = datetime.strptime(name, "%Y-%m-%d_%H-%M-%S")
        all_backups.append((name, dt, obj.size, obj.key))
    return sorted(all_backups, key=lambda x: x[1])


def get_backup_datestamp(url):
    jj = url.find("?")
    ii = url.rfind("/", 0, jj) + 1
    return url[ii:jj][:-4]


def get_secret(name, base64=False):
    try:
        value = os.environ[name]
    except KeyError:
        msg = f"Variable {Fore.RED}{name}{Fore.RESET} missing from environment."
        raise CommandError(msg)
    if base64:
        value = b64encode(value.encode()).decode()
    else:
        value = shlex.quote(value)
    return value


@contextmanager
def run_wrapper(intro):
    print(intro, end="", flush=True)
    try:
        yield
    except CommandError as e:
        print(f"{Fore.RED}error{Fore.RESET}")
        print(f" \u2514 {e}")
        sys.exit(1)
    except UnexpectedExit as e:
        print(f"{Fore.RED}error{Fore.RESET}")
        print(e)
        sys.exit(1)
    else:
        print(f"{Fore.GREEN}ok{Fore.RESET}")


def make_key(length=64):
    """Generate a sequence of random characters."""
    return "".join(
        random.SystemRandom().choice(string.ascii_uppercase + string.digits) for _ in range(length)
    )


def render_template(template: str, values: Dict, extra_labels: Dict) -> str:
    """Given a yaml of a K8s Job, replace template values and add extra labels to the pod spec"""
    extra_labels = extra_labels or {}
    values = values or {}

    # Replace keys using our templating language
    for k, v in values.items():
        template = template.replace("{{ %s }}" % k, v)

    job_json = yaml.safe_load(template)
    # Adding extra labels to k8s job pod spec
    for k, v in extra_labels.items():
        job_json["spec"]["template"]["metadata"]["labels"][k] = v
    return yaml.dump(job_json)


async def _run_job(
    path,
    values: Dict = None,
    context="",
    namespace="default",
    attach=False,
    cleanup=True,
    sequential=True,
    extra_labels: Dict = None,
):
    name = values["name"]
    logs = ""
    with tempfile.NamedTemporaryFile("wt", suffix=".yml") as tmp:
        # Generate yaml template to render
        rendered_template = render_template(
            open(APPS_PATH / ".." / path, "r").read(), values, extra_labels
        )
        tmp.write(rendered_template)
        tmp.flush()
        await async_run(f"kubectl create --context {context} -n {namespace} -f {tmp.name}")
        cmd = (
            "kubectl get pods"
            f" --context {context}"
            f" -n {namespace}"
            f" --selector=job-name={name}"
            " -o jsonpath='{.items[*].metadata.name}'"
        )

        @retry
        async def _find_pod():
            stdout, _, _ = await async_run(cmd)
            pod = stdout.decode()
            if not pod:
                raise CommandError("Failed to find pod.")
            return pod

        pod = await _find_pod()
        try:
            if not attach:
                intro = "Waiting for job to complete ... "
                with run_wrapper(intro):
                    cmd = (
                        "kubectl wait"
                        f" --context {context}"
                        f" -n {namespace}"
                        " --for=condition=complete"
                        " --timeout=-1s"
                        f" job/{name}"
                    )
                    _, _, output_log = await async_run(cmd)
                    logs += output_log
            else:
                await wait_for_pod(context, namespace, pod)
                cmd = f"kubectl attach --context {context} -n {namespace} -it {pod}"
                try:
                    if sequential:
                        # If we're not running asynchronously, use invoke.run to preserve interactivity for shell_plus, etc.
                        run(cmd, pty=True)
                    else:
                        _, _, output_log = await async_run(cmd)
                        logs += output_log

                except Exception as e:
                    if "current phase is Succeeded" not in str(e.result.stdout):
                        raise e
        except Exception as e:
            if "current phase is Succeeded" not in str(e.result.stdout):
                raise e
        finally:
            if cleanup:
                cmd = f"kubectl delete --context {context} -n {namespace} job {name}"
                await async_run(cmd)
    return logs


async def wait_for_pod(context, namespace, pod):
    while True:
        cmd = (
            "kubectl get pod"
            f" --context {context}"
            f" -n {namespace}"
            ' -o jsonpath="{.status.phase}"'
            f" {pod}"
        )
        stdout, _, _ = await async_run(cmd)
        stdout = stdout.decode().lower()
        if stdout != "pending":
            return stdout


def retry(*args, max_attempts=3, delay=1):
    """A decorator for retrying a function.

    After `max_attempts` failed attempts the last thrown exception
    is re-thrown.

    :param max_attempts: Maximum number of attempts.
    """

    def outer(f):
        if inspect.iscoroutinefunction(f):

            @wraps(f)
            async def inner(*args, **kwargs):
                attempt = 0
                while True:
                    try:
                        return await f(*args, **kwargs)
                    except Exception:
                        attempt += 1
                        if attempt >= max_attempts:
                            raise
                        await asyncio.sleep(delay)

        else:

            @wraps(f)
            def inner(*args, **kwargs):
                attempt = 0
                while True:
                    try:
                        return f(*args, **kwargs)
                    except Exception:
                        attempt += 1
                        if attempt >= max_attempts:
                            raise
                        time.sleep(delay)

        return inner

    if len(args) == 1 and callable(args[0]):
        return outer(args[0])
    else:
        return outer


def confirm_database(database):
    print(
        textwrap.fill(
            f"This operation will {Fore.RED}destroy{Fore.RESET}"
            f" the data currently stored in {Fore.RED}{database}{Fore.RESET}."
            " Please confirm by typing the name of the database"
            f" ({Fore.BLUE}{database}{Fore.RESET}):"
        )
    )
    value = input("> ")
    if value != database:
        print("Does not match, aborting.")
        sys.exit(1)



import asyncio
import base64
import os
import random

import dsnparse
from invoke import run, task

from .utils import kube
from .utils.apps import get_app_details
from .utils.cli import progress, warning

# TODO: Abstract this code out of gitops.


#####################
# DATABASE COMMANDS #
#####################
@task
def backup(ctx, app_name):
    """Backup production or staging database."""
    values = {"name": f"{app_name}-backup", "app": app_name}
    app = get_app_details(app_name, load_secrets=False)
    asyncio.run(
        kube._run_job(
            "jobs/backup-job.yml",
            values,
            context=app.cluster,
            namespace="workforce",
            sequential=True,
        )
    )


@task
def list_backups(ctx, app):
    kube.list_backups("workforce", app)


@task
def restore_backup(ctx, app_name, index, cleanup=True):
    """Restore backed up database."""
    kube.confirm_database(app_name)
    backups = kube.get_backups("workforce", app_name)
    backup = backups[int(index) - 1]
    values = {
        "name": f"{app_name}-restore",
        "timestamp": backup[0],
        "app": app_name,
    }
    app = get_app_details(app_name, load_secrets=False)
    asyncio.run(
        kube._run_job(
            "jobs/restore-job.yml",
            values,
            context=app.cluster,
            namespace="workforce",
            cleanup=cleanup,
        )
    )


@task
def copy_db(ctx, source, destination, skip_backup=False, cleanup=True):
    """Copy database between apps."""
    source_app = get_app_details(source, load_secrets=True)
    destination_app = get_app_details(destination, load_secrets=True)
    kube.confirm_database(destination)
    values = {
        "name": f"copy-db-{source}-{destination}",
        "source": source,
        "destination": destination,
        "SOURCE_DATABASE_URL_ENCODED": source_app.values["secrets"]["DATABASE_URL"],
        "DESTINATION_DATABASE_URL_ENCODED": destination_app.values["secrets"]["DATABASE_URL"],
        "skip_backup": "skip" if skip_backup else "",
    }
    asyncio.run(
        kube._run_job(
            "jobs/copy-db-job.yml",
            values,
            context=source_app.cluster,
            namespace="workforce",
            cleanup=cleanup,
        )
    )
    print(progress("You may want to clear the redis cache now!"))
    print(progress(f"\t- gitops mcommand {destination} clear_cache"))


@task
def download_backup(ctx, app, index=None, path=None, datestamp=False):
    """Download production or staging backup."""
    if not index:
        backups = kube.get_backups("workforce", app)
        if not backups:
            print(warning(f"No backups found for {app}"))
            return
        index = len(backups)
        print(progress("No index specified. Downloading latest backup: " + backups[-1][3]))

    kube.download_backup("workforce", app, index, path, datestamp=datestamp)


@task
def proxy(
    ctx, app_name, local_port=None, bastion_instance_id=None, aws_availability_zone=None, file=None
):
    """Creates a proxy to RDS. Can supply either the app name or a DSN

    Usage: gitops db.proxy app_name
    or     gitops db.proxy postgres://...:...@5432/db
           gitops db.proxy app_name --file=/tmp/address will write the proxy url to the file
    """
    try:
        database_url = app_name
        database_dsn = dsnparse.parse(database_url)
        app_name = database_dsn.user
    except ValueError:
        # app_name was not a valid dsn
        app = get_app_details(app_name, load_secrets=True)
        database_url = base64.b64decode(
            app.values["secrets"]["DATABASE_URL"].encode("ascii")
        ).decode("ascii")
        database_dsn = dsnparse.parse(database_url)

    if not local_port:
        local_port = random.randint(1000, 9999)

    modified_dsn = dsnparse.parse(database_url)
    modified_dsn.hostname = "localhost"
    modified_dsn.port = local_port

    bastion_instance_id = bastion_instance_id or os.environ.get("GITOPS_BASTION_INSTANCE_ID")
    if not bastion_instance_id:
        raise Exception(
            "Please set GITOPS_BASTION_INSTANCE_ID environment variable for db proxy to work."
        )

    aws_availability_zone = aws_availability_zone or os.environ.get("GITOPS_AWS_AVAILABILITY_ZONE")
    if not aws_availability_zone:
        raise Exception(
            "Please set GITOPS_AWS_AVAILABILITY_ZONE environment variable for db proxy to work."
        )

    print(progress(f"Creating a proxy to the RDS instance of: {app_name} "))

    # Create a temporary ssh key
    run("echo -e 'y\n' | ssh-keygen -t rsa -f /tmp/temp -N '' >/dev/null 2>&1")

    # Send ssh key to bastion. These only last 60 seconds
    run(
        f"""aws ec2-instance-connect send-ssh-public-key \
            --instance-id {bastion_instance_id}\
            --availability-zone  {aws_availability_zone} \
            --instance-os-user ec2-user \
            --ssh-public-key file:///tmp/temp.pub
    """,
        hide=True,
    )
    proxy_dsn = modified_dsn.geturl()
    print(progress(f"Connect to the db using: {proxy_dsn}\n"))
    if file:
        print(f"Outputing the proxy url to `{file}`")
        with open(file, "w") as fout:
            fout.write(proxy_dsn)
    # Create ssh tunnel
    cmd = f"""ssh -i /tmp/temp \
            -N -M -L {local_port}:{database_dsn.hostname}:{database_dsn.port} \
            -o "UserKnownHostsFile=/dev/null" \
            -o "StrictHostKeyChecking=no" \
            -o "ServerAliveInterval=60" \
            -o ProxyCommand="aws ssm start-session --target %h --document AWS-StartSSHSession --parameters portNumber=%p --region={aws_availability_zone[:-1]}" \
            ec2-user@{bastion_instance_id}
    """
    try:
        run(cmd, hide=True)
    finally:
        if file:
            os.remove(file)



import json
import random
import string
from configparser import ConfigParser

from invoke import run

CACHE = {}


def gen_secret(length=64):
    return "".join(
        random.SystemRandom().choice(string.ascii_uppercase + string.digits) for _ in range(length)
    )


def get_account_id():
    if "ACCOUNT_ID" not in CACHE:
        # This is not ideal, as it makes an assumption that the account_id of interest is the
        # one the user is currently sitting in. Ideally, should ask the cluster (though that is
        # messy in its own right, since we don't necessarily want the cluster that's in the
        # current context either).
        caller_identity = run("aws sts get-caller-identity", hide=True).stdout.strip()
        CACHE["ACCOUNT_ID"] = json.loads(caller_identity)["Account"]
    return CACHE["ACCOUNT_ID"]


class GitopsConfigParser(ConfigParser):
    def getlist(self, section, option, *args, **kwargs):
        value = self.get(section, option, *args, **kwargs)
        return list(filter(None, (x.strip() for x in value.splitlines())))


_config = GitopsConfigParser()
_config.read("setup.cfg")

if not _config.has_section("gitops"):
    _config.add_section("gitops")
config = _config["gitops"]



from contextlib import suppress
from hashlib import md5

import boto3
from colorama import Fore

from .cli import colourise

PREFIX_CACHE = {}
BATCH_SIZE = 100


def get_image(tag):
    """Finds a specific image in ECR."""
    # TODO
    raise NotImplementedError


def get_latest_image(prefix: str) -> str:
    """Finds latest image in ECR with the given prefix."""
    with suppress(KeyError):
        return PREFIX_CACHE[prefix]

    ecr = boto3.client("ecr")
    image_tags = []
    next_token = None
    while True:
        opts = {"repositoryName": "uptick", "filter": {"tagStatus": "TAGGED"}}
        if next_token:
            opts["nextToken"] = next_token
        results = ecr.list_images(**opts)
        for image in results["imageIds"]:
            if image["imageTag"].startswith(prefix + "-"):
                image_tags.append(image["imageTag"])
        next_token = results.get("nextToken")
        if not next_token:
            break

    if not image_tags:
        print(f'No images with tag "{prefix}-*".')
        PREFIX_CACHE[prefix] = None
        return None

    # ECR allows us to fetch 100 image details at a time.
    results = []
    for i in range(0, len(image_tags), BATCH_SIZE):
        batch_image_tags = image_tags[i : i + BATCH_SIZE]
        ecr_response = ecr.describe_images(
            repositoryName="uptick",
            imageIds=[{"imageTag": t} for t in batch_image_tags],
            filter={"tagStatus": "TAGGED"},
        )
        results += [(i["imagePushedAt"], i["imageTags"][0]) for i in ecr_response["imageDetails"]]

    latest_image_tag = sorted(results, key=lambda x: x[0], reverse=True)[0][1]
    PREFIX_CACHE[prefix] = latest_image_tag
    return latest_image_tag


def colour_image(image_tag: str):
    if not image_tag:
        return image_tag

    bits = image_tag.split("-")
    if len(bits) > 1:
        bits[0] = colourise(bits[0], color_hash(bits[1]))
        return "-".join(bits)
    else:
        return colourise(bits[0], color_hash(bits[0]))


def color_hash(bit):
    return [
        Fore.RED,
        Fore.GREEN,
        Fore.YELLOW,
        Fore.BLUE,
        Fore.MAGENTA,
        Fore.CYAN,
        Fore.WHITE,
    ][int.from_bytes(md5(bit.encode()).digest(), "big") % 7]



import asyncio

from invoke import task

from .core import command
from .utils.apps import get_app_details
from .utils.cli import success
from .utils.kube import run_job

UNSAFE_MANAGE_PY = "env DJANGO_ALLOW_ASYNC_UNSAFE=true ./manage.py"


@task
def bash(ctx, app, cleanup=True):
    """Brings up bash shell for selected app.

    eg. inv bash aesg
    """
    app = get_app_details(app)
    asyncio.run(run_job(app, "bash", cleanup=cleanup))
    print(success("Done!"))


@task
def mcommand(ctx, filter, mcommand, exclude="", cleanup=True, sequential=False, fargate=False):
    """Run django management command on selected app(s).

    eg. inv mcommand customer,sandbox -e aesg showmigrations
    """
    return command(
        ctx,
        filter,
        f"{UNSAFE_MANAGE_PY} {mcommand}",
        exclude=exclude,
        cleanup=cleanup,
        sequential=sequential,
        fargate=fargate,
    )


@task(aliases=["sp"])
def shell_plus(ctx, app, cleanup=True):
    """Brings up shell_plus for selected app.

    eg. inv sp aesg
    """
    app = get_app_details(app)
    asyncio.run(run_job(app, f"{UNSAFE_MANAGE_PY} shell_plus", cleanup=cleanup))
    print(success("Done!"))


@task
def migrate(
    ctx, filter, exclude="", cleanup=True, sequential=False, interactive=True, fargate=False
):
    """Runs migrations for selected app.

    eg. inv migrate workforce,sandbox
    """
    return command(
        ctx,
        filter,
        f"{UNSAFE_MANAGE_PY} migrate",
        exclude=exclude,
        cleanup=cleanup,
        sequential=sequential,
        interactive=interactive,
        fargate=fargate,
    )
